from dbus_next.service import ServiceInterface, dbus_property, method, signal, Variant
from dbus_next import PropertyAccess, RequestNameReply, BusType
from dbus_next.aio import MessageBus

import asyncio


async def setup_mpris(*names, bus_address=None, system=False):
    # TODO maybe they should all share a bus for speed
    async def setup(name):
        if system:
            bus_type = BusType.SYSTEM
        else:
            bus_type = BusType.SESSION
        bus = await MessageBus(bus_type=bus_type,
                               bus_address=bus_address).connect()
        player = MprisPlayer(bus)
        bus.export('/org/mpris/MediaPlayer2', player)
        bus.export('/org/mpris/MediaPlayer2', MprisRoot())
        reply = await bus.request_name(f'org.mpris.MediaPlayer2.{name}')
        assert reply == RequestNameReply.PRIMARY_OWNER
        return player

    players = await asyncio.gather(*(setup(name) for name in names))
    await asyncio.gather(*(p.ping() for p in players))
    return players


async def setup_playerctld(bus_address=None):
    bus = await MessageBus(bus_address=bus_address).connect()
    playerctld = PlayerctldInterface(bus)
    bus.export('/org/mpris/MediaPlayer2', playerctld)
    reply = await bus.request_name('org.mpris.MediaPlayer2.playerctld')
    assert reply == RequestNameReply.PRIMARY_OWNER
    return playerctld


class MprisRoot(ServiceInterface):
    def __init__(self):
        super().__init__('org.mpris.MediaPlayer2')

    @method()
    def Raise(self):
        return

    @method()
    def Quit(self):
        return

    @dbus_property(access=PropertyAccess.READ)
    def CanRaise(self) -> 'b':
        return False

    @dbus_property(access=PropertyAccess.READ)
    def HasTrackList(self) -> 'b':
        return False

    @dbus_property(access=PropertyAccess.READ)
    def Identity(self) -> 's':
        return 'playerctl test client'

    @dbus_property(access=PropertyAccess.READ)
    def SupportedUriSchemes(self) -> 'as':
        return ['file']

    @dbus_property(access=PropertyAccess.READ)
    def SupportedMimeTypes(self) -> 'as':
        return ['audio/mp3']


class MprisPlayer(ServiceInterface):
    def __init__(self, bus):
        super().__init__('org.mpris.MediaPlayer2.Player')
        self.counter = 0
        self.reset()
        self.bus = bus

    def reset(self):
        # method calls
        self.next_called = False
        self.previous_called = False
        self.pause_called = False
        self.play_pause_called = False
        self.stop_called = False
        self.play_called = False
        self.seek_called_with = None
        self.set_position_called_with = None
        self.open_uri_called_with = None

        # properties
        self.playback_status = 'Playing'
        self.loop_status = 'None'
        self.rate = 1.0
        self.shuffle = False
        self.metadata = {}
        self.volume = 1.0
        self.position = 0
        self.minimum_rate = 1.0
        self.maximum_rate = 1.0
        self.can_go_next = True
        self.can_go_previous = True
        self.can_play = True
        self.can_pause = True
        self.can_seek = True
        self.can_control = True

        # signals
        self.seeked_value = 0

    async def ping(self):
        await self.bus.introspect('org.freedesktop.DBus',
                                  '/org/freedesktop/DBus')

    async def set_artist_title(self, artist, title, track_id=None):
        if track_id is None:
            self.counter += 1
            track_id = '/' + str(self.counter)

        self.metadata = {
            'xesam:title': Variant('s', title),
            'xesam:artist': Variant('as', [artist]),
            'mpris:trackid': Variant('o', track_id),
        }

        self.emit_properties_changed({
            'Metadata': self.metadata,
        })
        await self.ping()

    async def clear_metadata(self):
        self.counter += 1
        self.metadata = {
            'mpris:trackid': Variant('o', '/' + str(self.counter)),
        }
        self.emit_properties_changed({
            'Metadata': self.metadata,
        })
        await self.ping()

    async def disconnect(self):
        self.bus.disconnect()
        await self.bus.wait_for_disconnect()

    @method()
    def Next(self):
        self.next_called = True

    @method()
    def Previous(self):
        self.previous_called = True

    @method()
    def Pause(self):
        self.pause_called = True

    @method()
    def PlayPause(self):
        self.play_pause_called = True

    @method()
    def Stop(self):
        self.stop_called = True

    @method()
    def Play(self):
        self.play_called = True

    @method()
    def Seek(self, offset: 'x'):
        self.seek_called_with = offset

    @method()
    def SetPosition(self, track_id: 'o', position: 'x'):
        self.set_position_called_with = (track_id, position)

    @method()
    def OpenUri(self, uri: 's'):
        self.open_uri_called_with = uri

    @signal()
    def Seeked(self) -> 'x':
        return self.seeked_value

    @dbus_property(access=PropertyAccess.READ)
    def PlaybackStatus(self) -> 's':
        return self.playback_status

    @dbus_property()
    def LoopStatus(self) -> 's':
        return self.loop_status

    @LoopStatus.setter
    def LoopStatus(self, status: 's'):
        self.loop_status = status

    @dbus_property()
    def Rate(self) -> 'd':
        return self.rate

    @Rate.setter
    def Rate(self, rate: 'd'):
        self.rate = rate

    @dbus_property()
    def Shuffle(self) -> 'b':
        return self.shuffle

    @Shuffle.setter
    def Shuffle(self, shuffle: 'b'):
        self.shuffle = shuffle

    @dbus_property(access=PropertyAccess.READ)
    def Metadata(self) -> 'a{sv}':
        return self.metadata

    @dbus_property()
    def Volume(self) -> 'd':
        return self.volume

    @Volume.setter
    def Volume(self, volume: 'd'):
        self.volume = volume

    @dbus_property(access=PropertyAccess.READ)
    def Position(self) -> 'x':
        return self.position

    @dbus_property(access=PropertyAccess.READ)
    def MinimumRate(self) -> 'd':
        return self.minimum_rate

    @dbus_property(access=PropertyAccess.READ)
    def MaximumRate(self) -> 'd':
        return self.maximum_rate

    @dbus_property(access=PropertyAccess.READ)
    def CanGoNext(self) -> 'b':
        return self.can_go_next

    @dbus_property(access=PropertyAccess.READ)
    def CanGoPrevious(self) -> 'b':
        return self.can_go_previous

    @dbus_property(access=PropertyAccess.READ)
    def CanPlay(self) -> 'b':
        return self.can_play

    @dbus_property(access=PropertyAccess.READ)
    def CanPause(self) -> 'b':
        return self.can_pause

    @dbus_property(access=PropertyAccess.READ)
    def CanSeek(self) -> 'b':
        return self.can_seek

    @dbus_property(access=PropertyAccess.READ)
    def CanControl(self) -> 'b':
        return self.can_control


class PlayerctldInterface(ServiceInterface):
    '''just enough of playerctld for testing'''
    def __init__(self, bus):
        super().__init__('com.github.altdesktop.playerctld')
        self.bus = bus
        self.player_names = []

    @dbus_property(access=PropertyAccess.READ)
    def PlayerNames(self) -> 'as':
        return self.player_names

    async def disconnect(self):
        self.bus.disconnect()
        await self.bus.wait_for_disconnect()



#!/usr/bin/env python3

from gi.repository import Playerctl, GLib

player = Playerctl.Player()


def on_metadata(player, metadata):
    if 'xesam:artist' in metadata.keys() and 'xesam:title' in metadata.keys():
        print('Now playing:')
        print('{artist} - {title}'.format(artist=metadata['xesam:artist'][0],
                                          title=metadata['xesam:title']))


def on_play(player, status):
    print('Playing at volume {}'.format(player.props.volume))


def on_pause(player, status):
    print('Paused the song: {}'.format(player.get_title()))


player.connect('playback-status::playing', on_play)
player.connect('playback-status::paused', on_pause)
player.connect('metadata', on_metadata)

# start playing some music
player.play()

if player.get_artist() == 'Lana Del Rey':
    player.next()

# wait for events
main = GLib.MainLoop()
main.run()



import asyncio
import os
from shlex import join


class CommandResult:
    def __init__(self, stdout, stderr, returncode):
        self.stdout = stdout.decode().strip()
        self.stderr = stderr.decode().strip()
        self.returncode = returncode


class PlayerctlProcess:
    def __init__(self, proc, debug=False):
        self.queue = asyncio.Queue()
        self.proc = proc

        async def reader(stream):
            while True:
                line = await stream.readline()
                if not line:
                    break
                line = line.decode().strip()
                if 'playerctl-DEBUG:' in line:
                    print(line)
                else:
                    self.queue.put_nowait(line)

        async def printer(stream):
            while True:
                line = await stream.readline()
                print(line)
                if not line:
                    break

        asyncio.get_event_loop().create_task(reader(proc.stdout))
        # asyncio.get_event_loop().create_task(printer(proc.stderr))

    def running(self):
        return self.proc.returncode is None


class PlayerctlCli:
    def __init__(self, bus_address=None, debug=False):
        self.bus_address = bus_address
        self.debug = debug
        self.proc = None

    async def _start(self, cmd):
        env = os.environ.copy()
        shell_cmd = f'playerctl {cmd}'

        if self.bus_address:
            env['DBUS_SESSION_BUS_ADDRESS'] = self.bus_address
        if self.debug:
            env['G_MESSAGES_DEBUG'] = 'playerctl'

        return await asyncio.create_subprocess_shell(
            shell_cmd,
            env=env,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)

    async def start(self, cmd):
        proc = await self._start(cmd)
        return PlayerctlProcess(proc)

    async def run(self, cmd):
        proc = await self._start(cmd)
        stdout, stderr = await proc.communicate()
        await proc.wait()
        return CommandResult(stdout, stderr, proc.returncode)

    async def list(self, players=[], ignored=[]):
        args = ['--list-all']

        if players:
            args.extend(['--player', ','.join(players)])

        if ignored:
            args.extend(['--ignored-players', ','.join(ignored)])

        cmd = await self.run(join(args))
        assert cmd.returncode == 0, cmd.stderr
        return cmd.stdout.splitlines()



#!/usr/bin/env python3

from gi.repository import GLib
import gi

gi.require_version('Playerctl', '2.0')
from gi.repository import Playerctl

manager = Playerctl.PlayerManager()

gi.require_version('Notify', '0.7')
from gi.repository import Notify

Notify.init("Media Player")
notification = Notify.Notification.new("")

from urllib.parse import urlparse, unquote
from pathlib import Path
from gi.repository import GdkPixbuf
import os.path


def notify(player):
    metadata = player.props.metadata
    keys = metadata.keys()
    if 'xesam:artist' in keys and 'xesam:title' in keys:
        notification.update(metadata['xesam:title'],
                            metadata['xesam:artist'][0])
        path = Path(unquote(urlparse(
            metadata['xesam:url']).path)).parent / "cover.jpg"
        if os.path.exists(path):
            image = GdkPixbuf.Pixbuf.new_from_file(str(path))
            notification.set_image_from_pixbuf(image)
        notification.show()


def on_play(player, status, manager):
    notify(player)


def on_metadata(player, metadata, manager):
    notify(player)


def init_player(name):
    player = Playerctl.Player.new_from_name(name)
    player.connect('playback-status::playing', on_play, manager)
    player.connect('metadata', on_metadata, manager)
    manager.manage_player(player)
    notify(player)


def on_name_appeared(manager, name):
    init_player(name)


manager.connect('name-appeared', on_name_appeared)

for name in manager.props.player_names:
    init_player(name)

main = GLib.MainLoop()
main.run()

Notify.uninit()



from dbus_next import Variant
from .mpris import setup_mpris
from .playerctl import PlayerctlCli

import pytest
import asyncio

# TODO: test missing function does not segv


@pytest.mark.asyncio
async def test_emoji(bus_address):
    [mpris] = await setup_mpris('emoji-format-test', bus_address=bus_address)
    mpris.metadata = {'mpris:length': Variant('x', 100000)}
    await mpris.ping()
    playerctl = PlayerctlCli(bus_address)

    status_emoji_cmd = 'metadata --format \'{{emoji(status)}}\''

    mpris.playback_status = 'Playing'
    cmd = await playerctl.run(status_emoji_cmd)
    assert cmd.stdout == '▶️', cmd.stderr

    mpris.playback_status = 'Paused'
    cmd = await playerctl.run(status_emoji_cmd)
    assert cmd.stdout == '⏸️', cmd.stderr

    mpris.playback_status = 'Stopped'
    cmd = await playerctl.run(status_emoji_cmd)
    assert cmd.stdout == '⏹️', cmd.stderr

    volume_emoji_cmd = 'metadata --format \'{{emoji(volume)}}\''
    mpris.volume = 0.0
    cmd = await playerctl.run(volume_emoji_cmd)
    assert cmd.stdout == '🔈', cmd.stderr

    mpris.volume = 0.5
    cmd = await playerctl.run(volume_emoji_cmd)
    assert cmd.stdout == '🔉', cmd.stderr

    mpris.volume = 1.0
    cmd = await playerctl.run(volume_emoji_cmd)
    assert cmd.stdout == '🔊', cmd.stderr

    cmd = await playerctl.run('metadata --format \'{{emoji("hi")}}\'')
    assert cmd.returncode == 1, cmd.stderr

    cmd = await playerctl.run('metadata --format \'{{emoji(status, volume)}}\''
                              )
    assert cmd.returncode == 1, cmd.stderr

    await mpris.disconnect()


class MetadataTest:
    def __init__(self, playerctl):
        self.tests = []
        self.playerctl = playerctl

    def add(self, fmt, expected, ret=0):
        fmt = fmt.replace("'", r"\'")
        self.tests.append((f"metadata --format '{fmt}'", expected, ret))

    async def run(self):
        coros = []
        for fmt, _, _ in self.tests:
            coros.append(self.playerctl.run(fmt))

        results = await asyncio.gather(*coros)

        for i, cmd in enumerate(results):
            fmt, expected, ret = self.tests[i]
            assert cmd.returncode == ret, cmd.stderr
            if ret == 0:
                assert cmd.stdout == expected, cmd.stderr


@pytest.mark.asyncio
async def test_format(bus_address):
    title = 'A Title'
    artist = 'An Artist'
    album = 'An Album'
    player_name = 'format-test'
    player_instance = f'{player_name}.instance123'

    [mpris] = await setup_mpris(player_instance, bus_address=bus_address)
    mpris.metadata = {
        'xesam:title': Variant('s', title),
        'xesam:artist': Variant('as', [artist]),
        'xesam:escapeme': Variant('s', '<hi>'),
        'xesam:album': Variant('s', album),
        'mpris:length': Variant('x', 100000)
    }
    mpris.volume = 2.0

    playerctl = PlayerctlCli(bus_address)

    test = MetadataTest(playerctl)

    test.add('{{artist}} - {{title}}', f'{artist} - {title}')
    test.add("{{markup_escape(xesam:escapeme)}}", "&lt;hi&gt;")
    test.add("{{lc(artist)}}", artist.lower())
    test.add("{{uc(title)}}", title.upper())
    test.add("{{uc(lc(title))}}", title.upper())
    test.add('{{uc("Hi")}}', "HI")
    test.add("{{mpris:length}}", "100000")
    test.add(
        '@{{ uc( "hi" ) }} - {{uc( lc( "HO"  ) ) }} . {{lc( uc(  title ) )   }}@',
        f'@HI - HO . {title.lower()}@')
    test.add("{{default(xesam:missing, artist)}}", artist)
    test.add("{{default(title, artist)}}", title)
    test.add('{{default("", "ok")}}', 'ok')
    test.add('{{default("ok", "not")}}', 'ok')
    test.add(' {{lc(album)}} ', album.lower())
    test.add('{{playerName}} - {{playerInstance}}',
             f'{player_name} - {player_instance}')

    await test.run()

    # numbers
    math = [
        '10',
        '-10 + 20',
        '10 + 10',
        '10 * 10',
        '10 / 10',
        '10 + 10 * 10 + 10',
        '10 + 10 * -10 + 10',
        '10 + 10 * -10 + -10',
        '-10 * 10 + 10',
        '-10 * -10 * -1 + -10',
        '-10 * 10 + -10 * -10 + 20 / 10 * -20 + -10',
        '8+-+--++-4',
        '2 - 10 * 1 + 1',
        '2 / -2 + 2 * 2 * -2 - 2 - 2 * -2',
        '2 * (2 + 2)',
        '10 * (10 + 12) - 4',
        '-(10)',
        '-(10 + 12 * -2)',
        '14 - (10 * 2 + 5) * -6',
        '(14 - 2 * 3) * (14 * -2 - 6) + -(4 - 2) * 5',
    ]

    # variables
    math += [
        'volume',
        'volume + 10',
        '-volume',
        '-volume * -1',
        '-volume + volume',
        'volume * volume',
        'volume * -volume',
        'volume + volume * -volume * volume + -volume',
        'volume / -volume + volume * volume * -volume - volume - volume * -volume',
        '-(volume + 3) * 5 * (volume + 2)',
    ]

    # functions
    math += [
        'default(5+5, None)',
        '-default(5 + 5, None)',
        '(-default(5 - 5, None) + 2) * 8',
        '2 + (5 * 4 + 3 * -default(5, default(6 * (3 + 4 * (6 + 2)) / 2, None)) + -56)',
    ]

    def default_shim(arg1, arg2):
        if arg1 is None:
            return arg2
        return arg1

    async def math_test(math):
        cmd = await playerctl.run("metadata --format '{{" + math + "}}'")
        assert cmd.returncode == 0, cmd.stderr
        assert float(cmd.stdout) == eval(math, {
            'volume': mpris.volume,
            'default': default_shim
        }), math

    await asyncio.gather(*[math_test(m) for m in math])

    await mpris.disconnect()



from .mpris import setup_mpris
from .playerctl import PlayerctlCli
import math

import asyncio
import pytest


@pytest.mark.asyncio
async def test_basics():
    playerctl = PlayerctlCli()

    result = await playerctl.run('--help')
    assert result.returncode == 0, result.stderr
    assert result.stdout
    assert not result.stderr

    # with no players
    result = await playerctl.run('--list-all')
    assert result.returncode == 0, result.stderr
    assert not result.stdout
    assert result.stderr

    result = await playerctl.run('--version')
    assert result.returncode == 0, result.stderr
    assert result.stdout
    assert not result.stderr

    commands = ('play', 'pause', 'play-pause', 'stop', 'next', 'previous',
                'position', 'position 5', 'volume', 'volume 0.5', 'status',
                'metadata', 'loop', 'loop None', 'shuffle', 'shuffle On',
                'open https://google.com')

    results = await asyncio.gather(*(playerctl.run(cmd) for cmd in commands))

    for result in results:
        assert result.returncode == 1
        assert not result.stdout
        assert 'No players found' in result.stderr.split('\n')


@pytest.mark.asyncio
async def test_list_names(bus_address):
    mpris_players = await setup_mpris('basics1',
                                      'basics2',
                                      'basics3',
                                      bus_address=bus_address)
    playerctl = PlayerctlCli(bus_address)

    result = await playerctl.run('--list-all')
    assert result.returncode == 0, result.stderr
    players = result.stdout.splitlines()
    assert 'basics1' in players
    assert 'basics2' in players
    assert 'basics3' in players

    await asyncio.gather(*[mpris.disconnect() for mpris in mpris_players])


@pytest.mark.asyncio
async def test_system_list_players(bus_address):
    system_players = await setup_mpris('system', system=True)
    session_players = await setup_mpris('session1', bus_address=bus_address)
    playerctl = PlayerctlCli(bus_address)
    result = await playerctl.run('-l')
    assert result.returncode == 0, result.stdout
    assert result.stdout.split() == ['session1', 'system']

    await asyncio.gather(
        *[mpris.disconnect() for mpris in system_players + session_players])


@pytest.mark.asyncio
async def test_queries(bus_address):
    [mpris] = await setup_mpris('queries', bus_address=bus_address)
    mpris.position = 2500000

    playerctl = PlayerctlCli(bus_address)

    query = await playerctl.run('status')
    assert query.stdout == mpris.playback_status, query.stderr

    query = await playerctl.run('volume')
    assert float(query.stdout) == mpris.volume, query.stderr

    query = await playerctl.run('loop')
    assert query.stdout == mpris.loop_status, query.stderr

    query = await playerctl.run('position')
    assert math.fabs(float(query.stdout) * 1000000 -
                     mpris.position) < 100, query.stderr

    query = await playerctl.run('shuffle')
    assert query.stdout == ('On' if mpris.shuffle else 'Off'), query.stderr



#!/usr/bin/env python3

from gi.repository import Playerctl, GLib

manager = Playerctl.PlayerManager()


def on_play(player, status, manager):
    print('player is playing: {}'.format(player.props.player_name))


def on_metadata(player, metadata, manager):
    keys = metadata.keys()
    if 'xesam:artist' in keys and 'xesam:title' in keys:
        print('{} - {}'.format(metadata['xesam:artist'][0],
                               metadata['xesam:title']))


def init_player(name):
    # choose if you want to manage the player based on the name
    if name.name in ['vlc', 'cmus']:
        player = Playerctl.Player.new_from_name(name)
        player.connect('playback-status::playing', on_play, manager)
        player.connect('metadata', on_metadata, manager)
        manager.manage_player(player)


def on_name_appeared(manager, name):
    init_player(name)


def on_player_vanished(manager, player):
    print('player has exited: {}'.format(player.props.player_name))


manager.connect('name-appeared', on_name_appeared)
manager.connect('player-vanished', on_player_vanished)

for name in manager.props.player_names:
    init_player(name)

main = GLib.MainLoop()
main.run()



from .mpris import setup_mpris, setup_playerctld
from .playerctl import PlayerctlCli
import pytest
import asyncio


def selector(bus_address):
    playerctl = PlayerctlCli(bus_address)

    async def select(*players):
        assert players
        cmd = '-p ' + str.join(
            ',', players) + ' status --format "{{playerInstance}}"'
        result = await playerctl.run(cmd)
        assert result.returncode == 0, result.stderr
        return result.stdout

    async def select_many(*players):
        assert players
        cmd = '--all-players -p ' + str.join(
            ',', players) + ' status --format "{{playerInstance}}"'
        result = await playerctl.run(cmd)
        assert result.returncode == 0, result.stderr
        return tuple(result.stdout.split('\n'))

    return select, select_many


@pytest.mark.asyncio
async def test_selection(bus_address):
    s1 = 'selection1'
    s1i = 'selection1.i_123'
    s2 = 'selection2'
    s3 = 'selection3'
    m4 = 'selection4'
    m5 = 'selection5'
    m6 = 'selection6'
    s6i = 'selection6.i_2'
    any_player = '%any'

    mpris_players = await setup_mpris(s1,
                                      s1i,
                                      s2,
                                      s3,
                                      s6i,
                                      bus_address=bus_address)

    # TODO: test ignored players
    selections = {
        (s1, ): (s1, s1i),
        (s3, s1): (s3, s1, s1i),
        (s2, s1, s3): (s2, s1, s1i, s3),
        (s1, s2): (s1, s1i, s2),
        (m4, m5, s2, s3): (s2, s3),
        (m5, s1, m4, s3): (s1, s1i, s3),
        (s1, s1i): (s1, s1i),
        (s1i, s1): (s1i, s1),
        (m6, s1): (s6i, s1, s1i),
        (m4, m6, s3): (s6i, s3),
        (any_player, ): (s1, s1i, s2, s3, s6i),
        (s1, any_player): (s1, s1i, s2, s3, s6i),  # s1 first
        (any_player, s1): (s2, s3, s6i, s1, s1i),  # s1 last
        (m6, any_player, s2): (s6i, s1, s1i, s3, s2),  # s6 first, s2 last
        (m6, s1, any_player, s2): (s6i, s1, s1i, s3, s2),
    }

    select, select_many = selector(bus_address)

    for selection, expected in selections.items():
        result = await select(*selection)
        assert result == expected[0], (selection, expected, result)
        result = await select_many(*selection)
        assert result == expected

    await asyncio.gather(*[mpris.disconnect() for mpris in mpris_players])


@pytest.mark.asyncio
async def test_daemon_selection(bus_address):
    playerctld = await setup_playerctld(bus_address=bus_address)
    playerctl = PlayerctlCli(bus_address)

    def iface_name(player_name):
        return f'org.mpris.MediaPlayer2.{player_name}'

    def set_players(players):
        playerctld.player_names = [iface_name(p) for p in players]

    s1 = 'selection1'
    s1i = 'selection1.i_123'
    s2 = 'selection2'
    s2i = 'selection2.i_123'
    s3 = 'selection3'
    m4 = 'selection4'
    m5 = 'selection5'
    m6 = 'selection6'
    s6i = 'selection6.i_2'
    any_player = '%any'

    # selection, players, expected result
    all_players = [s1, s1i, s2, s3, s6i]
    tests = [
        (None, all_players, all_players),
        (all_players, all_players, all_players),
        ([s2], [s1, s2], [s2]),
        ([s1], [s2, s1i, s1], [s1i, s1]),
        ([s1], [s2, s1, s1i], [s1, s1i]),
        ([s1i, s1], [s1, s1i], [s1i, s1]),
        ([any_player], all_players, all_players),
        ([any_player, s1], [s1, s1i, s2i, s2], [s2i, s2, s1, s1i]),
        ([any_player, s1], [s1, s1i, s2, s2i], [s2, s2i, s1, s1i]),
        ([any_player, s1], [s1i, s1, s2i, s2], [s2i, s2, s1i, s1]),
        ([any_player, s1], [s1i, s1, s2, s2i], [s2, s2i, s1i, s1]),
        ([s2, any_player], [s1, s1i, s2i, s2], [s2i, s2, s1, s1i]),
        ([s2, any_player], [s1, s1i, s2, s2i], [s2, s2i, s1, s1i]),
        ([s2, any_player], [s1i, s1, s2i, s2], [s2i, s2, s1i, s1]),
        ([s2, any_player], [s1i, s1, s2, s2i], [s2, s2i, s1i, s1]),
        ([s2i, any_player], [s1i, s1, s2, s2i], [s2i, s1i, s1, s2]),
    ]

    async def daemon_selection_test(test):
        selection, players, expected = test
        set_players(players)
        result = await playerctl.list(players=selection)
        assert result == expected, test

    for test in tests:
        # unfortunately it won't work in parallel because there's only one
        # playerctld
        await daemon_selection_test(test)

    await playerctld.disconnect()



import pytest
import asyncio


@pytest.fixture()
async def bus_address(scope='class'):
    proc = await asyncio.create_subprocess_shell(
        'dbus-launch', stdout=asyncio.subprocess.PIPE)
    stdout, __ = await proc.communicate()
    await proc.wait()
    assert proc.returncode == 0
    address = None
    for line in stdout.decode().split():
        if line.startswith('DBUS_SESSION_BUS_ADDRESS='):
            address = line.split('=', 1)[1].strip()
            break

    assert address

    return address



from .mpris import setup_mpris
from .playerctl import PlayerctlCli

import asyncio
import pytest

# TODO: test sending a command to all players


@pytest.mark.asyncio
async def test_commands(bus_address):
    [mpris] = await setup_mpris('commands', bus_address=bus_address)

    mpris.shuffle = False
    mpris.volume = 1.0
    mpris.loop_status = 'Track'

    commands = ('play', 'pause', 'play-pause', 'stop', 'next', 'previous')
    setters = ('volume 0.8', 'loop playlist', 'shuffle on')

    def get_called(cmd):
        return getattr(mpris, f'{cmd.replace("-", "_")}_called')

    playerctl = PlayerctlCli(bus_address)

    results = await asyncio.gather(*(playerctl.run(f'-p commands {cmd}')
                                     for cmd in commands + setters))

    for result in results:
        assert result.returncode == 0, result.stderr

    for i, cmd in enumerate(commands):
        result = results[i]
        assert get_called(cmd), f'{cmd} was not called: {result.stderr}'

    assert mpris.shuffle
    assert mpris.volume == 0.8
    assert mpris.loop_status == 'Playlist'

    await playerctl.run('-p commands shuffle toggle')

    assert not mpris.shuffle

    await mpris.disconnect()



from .mpris import setup_mpris
from .playerctl import PlayerctlCli
import pytest
import asyncio


@pytest.mark.asyncio
async def test_follow(bus_address):
    player1 = 'test1'
    [mpris1] = await setup_mpris(player1, bus_address=bus_address)

    playerctl = PlayerctlCli(bus_address)
    pctl_cmd = 'metadata --format "{{playerInstance}}: {{artist}} - {{title}}" --follow'
    proc = await playerctl.start(pctl_cmd)

    await mpris1.set_artist_title('artist', 'title')
    line = await proc.queue.get()
    assert line == 'test1: artist - title'

    await mpris1.set_artist_title('artist2', 'title2')
    line = await proc.queue.get()
    assert line == 'test1: artist2 - title2'

    await mpris1.clear_metadata()
    line = await proc.queue.get()
    assert line == 'test1:  -'

    await mpris1.set_artist_title('artist3', 'title3')
    line = await proc.queue.get()
    assert line == 'test1: artist3 - title3'

    await mpris1.disconnect()

    line = await proc.queue.get()
    assert line == ''


@pytest.mark.asyncio
async def test_follow_selection(bus_address):
    player1 = 'test1'
    player2 = 'test2'
    player3 = 'test3'
    player4 = 'test4'
    [mpris1, mpris2, mpris3,
     mpris4] = await setup_mpris(player1,
                                 player2,
                                 player3,
                                 player4,
                                 bus_address=bus_address)

    await mpris1.set_artist_title('artist', 'title')

    playerctl = PlayerctlCli(bus_address)
    pctl_cmd = '--player test3,test2,test1 metadata --format "{{playerInstance}}: {{artist}} - {{title}}" --follow'
    proc = await playerctl.start(pctl_cmd)

    line = await proc.queue.get()
    assert line == 'test1: artist - title'

    # player4 is ignored
    await mpris4.set_artist_title('artist', 'title')
    assert proc.queue.empty()

    # setting metadata the same twice doesn't print
    await mpris1.set_artist_title('artist', 'title')
    assert proc.queue.empty()

    await mpris2.set_artist_title('artist', 'title')
    line = await proc.queue.get()
    assert line == 'test2: artist - title'

    # player2 takes precedence
    await mpris1.set_artist_title('artist2', 'title2')
    assert proc.queue.empty()

    await mpris3.set_artist_title('artist', 'title')
    line = await proc.queue.get()
    assert line == 'test3: artist - title'

    # player 3 takes precedence
    await mpris2.set_artist_title('artist2', 'title2')
    assert proc.queue.empty()

    # when bus3 disconnects, it should show the next one
    await mpris3.disconnect()

    await mpris2.ping()
    line = await proc.queue.get()
    assert line == 'test2: artist2 - title2'

    # same for bus2
    await mpris2.disconnect()
    await mpris1.ping()
    line = await proc.queue.get()
    assert line == 'test1: artist2 - title2'

    await mpris1.disconnect()
    line = await proc.queue.get()
    assert line == ''

    await mpris4.disconnect()


@pytest.mark.asyncio
async def test_follow_selection_any(bus_address):
    player1 = 'test1'
    player2 = 'test2'
    player3 = 'test3'
    player4 = 'test4'
    [mpris1, mpris2, mpris3,
     mpris4] = await setup_mpris(player1,
                                 player2,
                                 player3,
                                 player4,
                                 bus_address=bus_address)

    playerctl = PlayerctlCli(bus_address)
    pctl_cmd = '--player test3,%any,test1 metadata --format "{{playerInstance}}: {{artist}} - {{title}}" --follow'
    proc = await playerctl.start(pctl_cmd)

    # test3 takes first precedence
    await mpris3.set_artist_title('artist', 'title')
    line = await proc.queue.get()
    assert line == 'test3: artist - title', proc.queue

    await mpris2.set_artist_title('artist', 'title')
    assert proc.queue.empty()

    await mpris1.set_artist_title('artist', 'title')
    assert proc.queue.empty()

    await mpris3.disconnect()
    line = await proc.queue.get()
    assert line == 'test2: artist - title'

    await mpris2.disconnect()
    line = await proc.queue.get()
    assert line == 'test1: artist - title'

    await mpris1.disconnect()
    line = await proc.queue.get()
    assert line == ''

    await mpris4.disconnect()


@pytest.mark.asyncio
async def test_follow_all_players(bus_address):
    player1 = 'test1'
    player2 = 'test2'
    player3 = 'test3'
    player4 = 'test4'
    [mpris1, mpris2, mpris3,
     mpris4] = await setup_mpris(player1,
                                 player2,
                                 player3,
                                 player4,
                                 bus_address=bus_address)

    await asyncio.gather(*[
        mpris.set_artist_title('artist', 'title')
        for mpris in [mpris1, mpris2, mpris3, mpris4]
    ])

    playerctl = PlayerctlCli(bus_address)
    pctl_cmd = '--all-players --player test3,test2,test1 metadata --format "{{playerInstance}}: {{artist}} - {{title}}" --follow'
    proc = await playerctl.start(pctl_cmd)

    # player4 is ignored
    await mpris4.set_artist_title('artist', 'title')
    assert proc.queue.empty()

    # no precedence, just whoever changes metadata last
    await mpris1.set_artist_title('artist2', 'title2')
    line = await proc.queue.get()
    assert line == 'test1: artist2 - title2'

    await mpris2.set_artist_title('artist2', 'title2')
    line = await proc.queue.get()
    assert line == 'test2: artist2 - title2'

    await mpris3.set_artist_title('artist2', 'title2')
    line = await proc.queue.get()
    assert line == 'test3: artist2 - title2'

    await mpris2.set_artist_title('artist2', 'title2')
    line = await proc.queue.get()
    assert line == 'test2: artist2 - title2'

    await mpris1.set_artist_title('artist2', 'title2')
    line = await proc.queue.get()
    assert line == 'test1: artist2 - title2'

    await mpris1.disconnect()
    await mpris4.ping()

    line = await proc.queue.get()
    assert line == 'test2: artist2 - title2'

    await mpris2.disconnect()
    await mpris4.ping()

    line = await proc.queue.get()
    assert line == 'test3: artist2 - title2'

    await mpris3.disconnect()
    await mpris4.ping()

    line = await proc.queue.get()
    assert line == ''

    await mpris4.disconnect()



import pytest

import os
from .mpris import setup_mpris
from .playerctl import PlayerctlCli
from dbus_next.aio import MessageBus
from dbus_next import Message, MessageType

import asyncio
from asyncio import Queue
from subprocess import run as run_process


async def start_playerctld(bus_address, debug=False):
    pkill = await asyncio.create_subprocess_shell('pkill playerctld')
    await pkill.wait()
    env = os.environ.copy()
    env['DBUS_SESSION_BUS_ADDRESS'] = bus_address
    env['G_MESSAGES_DEBUG'] = 'playerctl'
    proc = await asyncio.create_subprocess_shell(
        'playerctld',
        env=env,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT)

    async def printer(stream):
        while True:
            line = await stream.readline()
            print(line)
            if not line:
                break

    if debug:
        asyncio.get_event_loop().create_task(printer(proc.stdout))

    return proc


async def get_playerctld(bus):
    path = '/com/github/altdesktop/playerctld'
    interface = 'com.github.altdesktop.playerctld'
    introspection = await bus.introspect('org.mpris.MediaPlayer2.playerctld',
                                         path)
    obj = bus.get_proxy_object(interface, path, introspection)
    return obj.get_interface('org.freedesktop.DBus.Properties')


@pytest.mark.asyncio
async def test_daemon_commands(bus_address):
    playerctl = PlayerctlCli(bus_address)

    async def run(cmd):
        return await playerctl.run('-p playerctld ' + cmd)

    # with no other players running, these should error because there's no
    # active player (not no players found). This tests activation and property
    # errors as well.
    results = await asyncio.gather(*(run(cmd)
                                     for cmd in ('play', 'pause', 'play-pause',
                                                 'stop', 'next', 'previous',
                                                 'position', 'volume',
                                                 'status', 'metadata', 'loop',
                                                 'shuffle')))
    for result in results:
        assert result.returncode == 1
        assert 'No player could handle this command' in result.stderr.splitlines(
        )

    # restart playerctld so we can manage the process and see debug info
    playerctld_proc = await start_playerctld(bus_address)

    [mpris1, mpris2, mpris3] = await setup_mpris('daemon1',
                                                 'daemon2',
                                                 'daemon3',
                                                 bus_address=bus_address)
    await mpris2.set_artist_title('artist', 'title')
    cmd = await run('play')
    assert cmd.returncode == 0, cmd.stdout
    assert mpris2.play_called, cmd.stdout
    mpris2.reset()

    await mpris1.set_artist_title('artist', 'title')
    cmd = await run('play')
    assert cmd.returncode == 0, cmd.stderr
    assert mpris1.play_called
    mpris1.reset()

    await mpris3.set_artist_title('artist', 'title')
    cmd = await run('play')
    assert cmd.returncode == 0, cmd.stderr
    assert mpris3.play_called
    mpris3.reset()

    await mpris3.disconnect()
    cmd = await run('play')
    assert cmd.returncode == 0, cmd.stderr
    assert mpris1.play_called
    mpris1.reset()

    await asyncio.gather(mpris1.disconnect(), mpris2.disconnect())

    playerctld_proc.terminate()
    await playerctld_proc.wait()


@pytest.mark.asyncio
async def test_daemon_follow(bus_address):
    playerctld_proc = await start_playerctld(bus_address)

    [mpris1, mpris2] = await setup_mpris('player1',
                                         'player2',
                                         bus_address=bus_address)
    playerctl = PlayerctlCli(bus_address)
    pctl_cmd = '--player playerctld metadata --format "{{playerInstance}}: {{artist}} - {{title}}" --follow'
    proc = await playerctl.start(pctl_cmd)

    await mpris1.set_artist_title('artist1', 'title1')
    line = await proc.queue.get()
    assert line == 'playerctld: artist1 - title1', proc.queue

    await mpris2.set_artist_title('artist2', 'title2')
    line = await proc.queue.get()
    assert line == 'playerctld: artist2 - title2', proc.queue

    [mpris3] = await setup_mpris('player3', bus_address=bus_address)
    await mpris3.set_artist_title('artist3', 'title3')
    line = await proc.queue.get()

    if line == '':
        # the line might be blank here because of the test setup
        line = await proc.queue.get()

    assert line == 'playerctld: artist3 - title3', proc.queue

    await mpris1.set_artist_title('artist4', 'title4')
    line = await proc.queue.get()
    assert line == 'playerctld: artist4 - title4', proc.queue

    await mpris1.set_artist_title('artist5', 'title5')
    line = await proc.queue.get()
    assert line == 'playerctld: artist5 - title5', proc.queue

    await mpris1.disconnect()
    line = await proc.queue.get()
    assert line == 'playerctld: artist3 - title3', proc.queue

    await asyncio.gather(mpris2.disconnect(), mpris3.disconnect())

    playerctld_proc.terminate()
    proc.proc.terminate()
    await proc.proc.wait()
    await playerctld_proc.wait()


async def playerctld_shift(bus_address, reverse=False):
    env = os.environ.copy()
    env['DBUS_SESSION_BUS_ADDRESS'] = bus_address
    env['G_MESSAGES_DEBUG'] = 'playerctl'
    cmd = 'playerctld unshift' if reverse else 'playerctld shift'
    shift = await asyncio.create_subprocess_shell(
        cmd,
        env=env,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT)
    return await shift.wait()


@pytest.mark.asyncio
async def test_daemon_shift_simple(bus_address):
    playerctld_proc = await start_playerctld(bus_address)

    mprises = await setup_mpris('player1',
                                'player2',
                                'player3',
                                bus_address=bus_address)
    [mpris1, mpris2, mpris3] = mprises

    playerctl = PlayerctlCli(bus_address)
    pctl_cmd = '--player playerctld metadata --format "{{playerInstance}}: {{artist}} - {{title}}" --follow'
    proc = await playerctl.start(pctl_cmd)

    await mpris1.set_artist_title('artist1', 'title1')
    line = await proc.queue.get()
    assert line == 'playerctld: artist1 - title1', proc.queue

    await mpris2.set_artist_title('artist2', 'title2')
    line = await proc.queue.get()
    assert line == 'playerctld: artist2 - title2', proc.queue

    await mpris3.set_artist_title('artist3', 'title3')
    line = await proc.queue.get()
    assert line == 'playerctld: artist3 - title3', proc.queue

    code = await playerctld_shift(bus_address)
    assert code == 0
    line = await proc.queue.get()
    assert line == 'playerctld: artist2 - title2', proc.queue

    code = await playerctld_shift(bus_address)
    assert code == 0
    line = await proc.queue.get()
    assert line == 'playerctld: artist1 - title1', proc.queue

    code = await playerctld_shift(bus_address, reverse=True)
    assert code == 0
    line = await proc.queue.get()
    assert line == 'playerctld: artist2 - title2', proc.queue

    code = await playerctld_shift(bus_address, reverse=True)
    assert code == 0
    line = await proc.queue.get()
    assert line == 'playerctld: artist3 - title3', proc.queue

    playerctld_proc.terminate()
    proc.proc.terminate()
    await asyncio.gather(mpris1.disconnect(), mpris2.disconnect(),
                         playerctld_proc.wait(), proc.proc.wait())


@pytest.mark.asyncio
async def test_daemon_shift_no_player(bus_address):
    playerctld_proc = await start_playerctld(bus_address)

    playerctl = PlayerctlCli(bus_address)
    pctl_cmd = '--player playerctld metadata --format "{{playerInstance}}: {{artist}} - {{title}}" --follow'
    proc = await playerctl.start(pctl_cmd)

    code = await playerctld_shift(bus_address)
    assert code == 1

    [mpris1] = await setup_mpris('player1', bus_address=bus_address)
    code = await playerctld_shift(bus_address)
    assert code == 0

    await mpris1.disconnect()
    code = await playerctld_shift(bus_address)
    assert code == 1

    code = await playerctld_shift(bus_address, reverse=True)
    assert code == 1

    [mpris1] = await setup_mpris('player1', bus_address=bus_address)
    code = await playerctld_shift(bus_address, reverse=True)
    assert code == 0

    await mpris1.disconnect()
    code = await playerctld_shift(bus_address, reverse=True)
    assert code == 1

    playerctld_proc.terminate()
    await playerctld_proc.wait()


@pytest.mark.asyncio
async def test_active_player_change(bus_address):
    queue = Queue()
    playerctld_proc = await start_playerctld(bus_address)

    bus = await MessageBus(bus_address=bus_address).connect()

    reply = await bus.call(
        Message(destination='org.freedesktop.DBus',
                interface='org.freedesktop.DBus',
                path='/org/freedesktop/DBus',
                member='AddMatch',
                signature='s',
                body=["sender='org.mpris.MediaPlayer2.playerctld'"]))

    assert reply.message_type == MessageType.METHOD_RETURN, reply.body

    def message_handler(message):
        if message.member == 'PropertiesChanged' and message.body[
                0] == 'com.github.altdesktop.playerctld' and 'PlayerNames' in message.body[
                    1]:
            queue.put_nowait(message.body[1]['PlayerNames'].value)

    def player_list(*args):
        return [f'org.mpris.MediaPlayer2.{name}' for name in args]

    bus.add_message_handler(message_handler)

    [mpris1] = await setup_mpris('player1', bus_address=bus_address)

    assert player_list('player1') == await queue.get()

    [mpris2] = await setup_mpris('player2', bus_address=bus_address)

    assert player_list('player2', 'player1') == await queue.get()

    # changing artist/title should bump the player up
    await mpris1.set_artist_title('artist1', 'title1', '/1')

    assert player_list('player1', 'player2') == await queue.get()

    # if properties are not actually different, it shouldn't update
    await mpris2.set_artist_title('artist2', 'title2', '/2')
    assert player_list('player2', 'player1') == await queue.get()

    await mpris1.set_artist_title('artist1', 'title1', '/1')
    await mpris1.ping()
    assert queue.empty()

    bus.disconnect()
    await asyncio.gather(mpris1.disconnect(), mpris2.disconnect(),
                         bus.wait_for_disconnect())

    playerctld_proc.terminate()
    await playerctld_proc.wait()
